DungeonMate = LibStub("AceAddon-3.0"):NewAddon("DungeonMate","AceConsole-3.0","AceEvent-3.0")
local lgt = LibStub("LibGroupTalents-1.0")

local version = 0.2

local dm_panel, plist

local ignorelist = {} -- Periodically updated list of every ignore on the character. UPDATE BEFORE USING.

-- prototyped functions
local clearParty, insertPlayer, insertInfoline, clearAllLines, UpdatePartyDisplay

local roleicons = "Interface\\LFGFrame\\UI-LFG-ICON-PORTRAITROLES.blp"
local dmicon = "Interface\\Icons\\Spell_Nature_StormReach"

local function SetPlayerIgnoreFavoriteIndicators(frame,isIgnored,isFavorite)
	frame:SetBackdropColor(0,0,0,0.8)

	if isFavorite then
		frame:SetBackdropColor(0,0,1,0.3)
		--frame.linebox.buttonbar.favoritebutton:SetText("Remove Favorite")
	end
	
	if isIgnored then
		frame:SetBackdropColor(1,0,0,0.3)
		frame.linebox.buttonbar.ignorebutton:SetText("Un-Ignore")
	end
end

local function UpdateIgnorelist()
	ignorelist = {} -- flush out the old
	local i = GetNumIgnores()
	while i > 0 do
		ignorelist[GetIgnoreName(i)]=true
		i = i - 1
	end
end

local function GenerateGroupInfo()
	local result = {}
	local n = GetNumPartyMembers()
	while n > 0 do
		local unit = "party" .. n
		local name = GetUnitName(unit,true)
		if name ~= "Unknown" then
			local isTank, isHeal, isDPS = UnitGroupRolesAssigned(unit)
			local classname,class = UnitClass(unit)
			local level = UnitLevel(unit)
			local dominant, t1, t2, t3 = lgt:GetUnitTalentSpec(unit)
			result[name] = {["name"] = name} -- Yeah. I know. Weird, huh?
			if isTank then
				result[name].role = "tank"
			elseif isHeal then
				result[name].role = "heal"
			elseif isDPS then
				result[name].role = "dps"
			else
				result[name].role = "no role"
			end
			result[name].class = class
			result[name].classname = classname
			result[name].level = level
			if dominant then
				result[name].spec = {
					["dominant"] = dominant,
					["t1"] = t1,
					["t2"] = t2,
					["t3"] = t3,
				}
			end
			
			-- Favorite --
			result[name].isFavorite = false -- false for now!
			
			-- Ignored --
			if ignorelist[name] then -- I use this construct to ensure only 'true' or 'false' gets returned
				result[name].isIgnored = true
			else
				result[name].isIgnored = false
			end
			
			-- GearGauge --
			if GearGauge then
				local rating, equip_level = GearGauge:Scan(unit)
				result[name].GearGuage = {
					["rating"] = rating or "??",
					["equip_level"] = equip_level or "??"
				}
			end
			-- End GearGauge --
		end
		n = n - 1
	end
	return result
end

local function updatePlayer(player, frame)
	frame.info = player -- export the original info, in case it helps anyone.
	
	SetPlayerIgnoreFavoriteIndicators(frame,player.isIgnored,player.isFavorite)
		
	
	if player.role == "tank" then
		frame.titlebar.rolebutton.texture:SetTexture(roleicons)
		frame.titlebar.rolebutton.texture:SetTexCoord(0, 19/64, 22/64, 41/64)
		frame.titlebar.rolebutton.texture:Show()
	elseif player.role == "heal" then
		frame.titlebar.rolebutton.texture:SetTexture(roleicons)
		frame.titlebar.rolebutton.texture:SetTexCoord(20/64, 39/64, 1/64, 20/64)
		frame.titlebar.rolebutton.texture:Show()
	elseif player.role == "dps" then
		frame.titlebar.rolebutton.texture:SetTexture(roleicons)
		frame.titlebar.rolebutton.texture:SetTexCoord(20/64, 39/64, 22/64, 41/64)
		frame.titlebar.rolebutton.texture:Show()
	else
		frame.titlebar.rolebutton.texture:SetTexture(dmicon)
		frame.titlebar.rolebutton.texture:SetTexCoord(0,1,0,1)
		frame.titlebar.rolebutton.texture:Show()
	end
	
	clearAllLines(frame)
	
	local classinfo = player.level .. " " .. player.classname
	
	if player.spec then
		classinfo = classinfo.." "..player.spec.t1.."/"..player.spec.t2.."/"..player.spec.t3
		classinfo = classinfo .. " " .. player.spec.dominant 
	end
	
	insertInfoline(classinfo, frame)
	
	-- GearGauge --
	if player.GearGauge then
		insertInfoline("GearGuage   Rating: "..player.GearGauge.rating .."iLevel: " .. player.GearGauge.equip_level)
	end

end

local function updateStalePlayer(frame)
	-- This function is set up to update players which are no longer part of the group. 
	-- It is an analogue to updatePlayer, with the benefit of not needing to update most values.
	-- Some tasks:
	-- 1) Change the role icon to the no-role icon.
	-- 2) Check for (and update to) ignore and favorite statii.
	-- IMPORTANT NOTE! It is not generally safe to assume that frame.info is accurate, as this function should
	-- update that value! (Some values, like frame.info.name, are perfectly safe.)
	frame.info.isFavorite = false -- For now!
	frame.info.isIgnored = ignorelist[frame.info.name]
	
	-- Role Icon
	frame.titlebar.rolebutton.texture:SetTexture(dmicon)
	frame.titlebar.rolebutton.texture:SetTexCoord(0,1,0,1)
	frame.titlebar.rolebutton.texture:Show()
	
	-- Favorite/Ignore update
	SetPlayerIgnoreFavoriteIndicators(frame,frame.info.isIgnored,frame.info.isFavorite)
	
	-- Note: We do NOT clear all lines, as we have no reliable way of reconstructing them!
	
end

-- PROTOTYPED as local
function UpdatePartyDisplay()
	if not plist then return end
	local group = GenerateGroupInfo()
	UpdateIgnorelist()
	local v
	
	for _,v in ipairs(plist.group) do
		if group[v.info.name] then
			-- This frame belongs to a current member of the party, so use the GroupInfo
			updatePlayer(group[v.info.name],v)
			group[v.info.name].found = true
		else
			-- This frame is "stale", but should still be updated for ignore, notes, etc.
			updateStalePlayer(v)
		end
	end
	-- Now, add all the group members that haven't been seen before
	for _,v in pairs(group) do
		if not v.found then
			insertPlayer(v)
		end
	end
end

local function ResetDisplay()
	clearParty()
	UpdatePartyDisplay()
end

local function showPanel()
	UpdatePartyDisplay()
	dm_panel:Show()
end

local function hidePanel()
	dm_panel:Hide()
end

local function saveFramePositions()
		local point,rt,rp,xoff,yoff = plist:GetPoint(1)
		DungeonMate.db.char.panel_position.x = xoff
		DungeonMate.db.char.panel_position.y = yoff
		DungeonMate.db.char.panel_position.point = point
end

local function initializeFrame()
	dm_panel = CreateFrame("Frame",nil,UIParent)
	dm_panel:SetBackdrop({bgFile ="Interface/DialogFrame/UI-DialogBox-Background", 
						  edgeFile = "Interface/Tooltips/UI-Tooltip-Border", 
						  tile = true, tileSize = 16, edgeSize = 16, 
						  insets = { left = 4, right = 4, top = 4, bottom = 4 }})

	
	local title = dm_panel:CreateFontString(nil,"OVERLAY","GameFontHighlightLarge")
	title:SetFont("Fonts\\FRIZQT__.TTF", 20, "OUTLINE")
	title:SetPoint("TOPLEFT",dm_panel,15,-11)
	title:SetText("DungeonMate v" .. version)
	title:Show()
	
	local footer = CreateFrame("Frame",nil,dm_panel)
	footer:SetHeight(25)
	footer:Show()
	
	-- This button used to say "Update", hence the name. 
	local update = CreateFrame("Button",nil,footer,"UIPanelButtonTemplate")
	update:SetText("Reset")
	update:SetScript("OnClick",ResetDisplay)
	update:Show()
	
	local close = CreateFrame("Button",nil,dm_panel,"UIPanelCloseButton")
	close:SetPoint("TOPRIGHT",dm_panel,"TOPRIGHT")
	close:SetScript("OnClick",hidePanel)

	plist = CreateFrame("Frame",nil,dm_panel)
	plist:SetPoint(DungeonMate.db.char.panel_position.point,UIParent,
	               DungeonMate.db.char.panel_position.point,
	               DungeonMate.db.char.panel_position.x,
	               DungeonMate.db.char.panel_position.y)
	plist:SetWidth(250)
	plist:SetHeight(4) -- dummy height, changes with added players
	plist.group = {}
	plist:Show()
	
	
	footer:SetPoint("BOTTOMLEFT",plist,"BOTTOMLEFT",0,-1*(footer:GetHeight()+4))
	footer:SetPoint("BOTTOMRIGHT",plist,"BOTTOMRIGHT",0,-1*(footer:GetHeight()+4))
	update:SetPoint("TOPLEFT",footer,"TOPLEFT",50,0)
	update:SetPoint("BOTTOMRIGHT",footer,"BOTTOMRIGHT",-50,0)
	dm_panel:SetPoint("TOPLEFT",plist,"TOPLEFT",-10,30)
	dm_panel:SetPoint("BOTTOMRIGHT",footer,"BOTTOMRIGHT",10,-5)
	
	dm_panel:EnableMouse(true)
	plist:SetMovable(true)
	dm_panel:SetScript("OnMouseUp",function(frame)
		plist:StopMovingOrSizing()
		saveFramePositions()
	end)
	dm_panel:SetScript("OnMouseDown",function(frame,button)
		if button == "LeftButton" then
			plist:StartMoving()
		end
	end)
	dm_panel:SetScript("OnHide",function(frame)
		plist:StopMovingOrSizing()
		saveFramePositions()
	end)
	
	dm_panel:Show()
end

local infolineRecycleQueue = {}

local function recycleInfoline(line)
	line:Hide()
	tinsert(infolineRecycleQueue,line)
end

local function updateInfoline(frame,text)
	frame.label:SetText(text)
end


local function manufactureInfoline(text,parent)
	local resultframe
	if #infolineRecycleQueue > 0 then
		resultframe = tremove(infolineRecycleQueue)
	else
		resultframe = CreateFrame("Frame",nil,parent)
		resultframe.label = resultframe:CreateFontString(nil,"OVERLAY","GameFontNormal")
		resultframe.label:SetFont("Fonts\\FRIZQT__.TTF", 12)
		resultframe.label:SetPoint("LEFT",resultframe,"LEFT",0,0)
		resultframe:SetHeight(15)

	end
	updateInfoline(resultframe,text)
	resultframe:Show()
	return resultframe
end

-- PROTOTYPED as local
function insertInfoline(text, playerframe)
	local infoline = manufactureInfoline(text, playerframe.linebox)
	if #(playerframe.linebox.lines) > 0 then
		local parent = playerframe.linebox.lines[#(playerframe.linebox.lines)]
		infoline:SetPoint("TOPLEFT",parent,"TOPLEFT")
		infoline:SetPoint("TOPRIGHT",parent,"TOPRIGHT")
	else
		infoline:SetPoint("TOPLEFT",playerframe.linebox,"TOPLEFT")
		infoline:SetPoint("TOPRIGHT",playerframe.linebox,"TOPRIGHT")
	end
	playerframe.linebox:SetHeight(playerframe.linebox:GetHeight()+infoline:GetHeight())
	playerframe:SetHeight(playerframe:GetHeight()+infoline:GetHeight())
	tinsert(playerframe.linebox.lines,infoline)
end

-- PROTOTYPED as local
function clearAllLines(playerframe)
	while #(playerframe.linebox.lines) > 0 do
		local temp = tremove(playerframe.linebox.lines)
		recycleInfoline(temp)
		playerframe.linebox:SetHeight(playerframe.linebox:GetHeight()-temp:GetHeight())
		playerframe:SetHeight(playerframe:GetHeight()-temp:GetHeight())
	end
end

local playerRecycleQueue = {}

local function recyclePlayer(player)
	player:Hide()
	clearAllLines(player)
	tinsert(playerRecycleQueue,player)
end

local function manufacturePlayerFrame(player)
	local resultframe -- Input player is a player info table. Output is a frame based on that.
	if #playerRecycleQueue > 0 then
		resultframe = tremove(playerRecycleQueue) -- ok, so it's a stack... big deal
	else
		resultframe = CreateFrame("Frame",nil,plist)
		resultframe:SetHeight(45) -- filled by the titlebar + buttonbar
		resultframe:SetBackdrop({ bgFile ="Interface\\Addons\\DungeonMate\\blank", 
		                          edgeFile = "Interface/Tooltips/UI-Tooltip-Border", 
		                          tile = true, tileSize = 16, edgeSize = 16, 
		                          insets = { left = 4, right = 4, top = 4, bottom = 4 }})

		
		resultframe.titlebar = CreateFrame("Frame",nil,resultframe)
		resultframe.titlebar:SetPoint("TOPLEFT",resultframe,"TOPLEFT")
		resultframe.titlebar:SetPoint("TOPRIGHT",resultframe,"TOPRIGHT")
		resultframe.titlebar:SetHeight(20)
		resultframe.titlebar:Show()
		
		
		resultframe.titlebar.label = resultframe.titlebar:CreateFontString(nil,"OVERLAY","GameFontNormal")
		resultframe.titlebar.label:SetFont("Fonts\\FRIZQT__.TTF", 14)
		resultframe.titlebar.label:SetPoint("TOPLEFT",resultframe,"TOPLEFT",25,-5)
		
		resultframe.titlebar.rolebutton = CreateFrame("Button",nil,resultframe.titlebar)
		resultframe.titlebar.rolebutton:SetPoint("TOPLEFT",resultframe,"TOPLEFT",5,-5)
		resultframe.titlebar.rolebutton:SetHeight(16)
		resultframe.titlebar.rolebutton:SetWidth(16)
		resultframe.titlebar.rolebutton:EnableMouse(true)


		resultframe.titlebar.rolebutton.texture = resultframe.titlebar.rolebutton:CreateTexture(nil,"OVERLAY")
		resultframe.titlebar.rolebutton.texture:SetAllPoints(resultframe.titlebar.rolebutton)
		
		resultframe.linebox = CreateFrame("Frame",nil,resultframe)
		resultframe.linebox:SetPoint("TOPLEFT",resultframe.titlebar,"BOTTOMLEFT",25,0)
		resultframe.linebox:SetPoint("TOPRIGHT",resultframe.titlebar,"BOTTOMRIGHT",0,0)
		resultframe.linebox:SetHeight(20) -- filled up by the buttonbar below
		
		resultframe.linebox:Show()
		
		resultframe.linebox.buttonbar = CreateFrame("Frame",nil,resultframe.linebox)
		resultframe.linebox.buttonbar:SetPoint("BOTTOMLEFT",resultframe.linebox,"BOTTOMLEFT")
		resultframe.linebox.buttonbar:SetPoint("BOTTOMRIGHT",resultframe.linebox,"BOTTOMRIGHT")
		resultframe.linebox.buttonbar:SetHeight(20) -- fills up the default space of the linebox
		resultframe.linebox.buttonbar:Show()
		
		resultframe.linebox.buttonbar.ignorebutton = CreateFrame("Button",nil,resultframe.linebox.buttonbar,"UIPanelButtonTemplate")
		resultframe.linebox.buttonbar.ignorebutton:SetPoint("TOPLEFT",resultframe.linebox.buttonbar,"TOPLEFT")
		resultframe.linebox.buttonbar.ignorebutton:SetPoint("BOTTOMLEFT",resultframe.linebox.buttonbar,"BOTTOMLEFT")
		resultframe.linebox.buttonbar.ignorebutton:SetWidth(60)
		resultframe.linebox.buttonbar.ignorebutton:SetText("Ignore")
		resultframe.linebox.buttonbar.ignorebutton:GetFontString():SetFont("Fonts\\FRIZQT__.TTF", 10)
		resultframe.linebox.buttonbar.ignorebutton:GetFontString():SetPoint("CENTER",resultframe.linebox.buttonbar.ignorebutton,"CENTER",0,1)
		resultframe.linebox.buttonbar.ignorebutton:Show();
		
		resultframe.linebox.lines = {}
		
		resultframe.titlebar.rolebutton:SetScript("OnClick",function(frame)
			if resultframe.linebox:IsShown() then
				resultframe:SetHeight(resultframe:GetHeight()-resultframe.linebox:GetHeight())
				plist:SetHeight(plist:GetHeight()-resultframe.linebox:GetHeight())
				resultframe.linebox:Hide()
			else
				resultframe:SetHeight(resultframe:GetHeight()+resultframe.linebox:GetHeight())
				plist:SetHeight(plist:GetHeight()+resultframe.linebox:GetHeight())
				resultframe.linebox:Show()
			end
		end)
	end
	
	resultframe.titlebar.label:SetText(player.name)
	resultframe.linebox.buttonbar.ignorebutton:SetScript("OnClick",function(frame) AddOrDelIgnore(player.name) end)	
	
	if player.class == "DEATHKNIGHT" then
		frame.titlebar.label:SetTextColor(0.77,0.12,0.23)
	elseif player.class == "DRUID" then
		frame.titlebar.label:SetTextColor(1.00,0.49,0.04)
	elseif player.class == "HUNTER" then
		frame.titlebar.label:SetTextColor(0.67,0.83,0.45)
	elseif player.class == "MAGE" then
		frame.titlebar.label:SetTextColor(0.41,0.80,0.94)
	elseif player.class == "PALADIN" then
		frame.titlebar.label:SetTextColor(0.96,0.55,0.73)
	elseif player.class == "PRIEST" then
		frame.titlebar.label:SetTextColor(1.00,1.00,1.00)
	elseif player.class == "ROGUE" then
		frame.titlebar.label:SetTextColor(1.00,0.96,0.41)
	elseif player.class == "SHAMAN" then
		frame.titlebar.label:SetTextColor(0.14,0.35,1.00)
	elseif player.class == "WARLOCK" then
		frame.titlebar.label:SetTextColor(0.58,0.51,0.79)
	elseif player.class == "WARRIOR" then
		frame.titlebar.label:SetTextColor(0.78,0.61,0.43)
	else 
		frame.titlebar.label:SetTextColor(0.5,0.5,0.5)
	end
	
	updatePlayer(player,resultframe)
	
	resultframe:Show()
	return resultframe
end

-- PROTOTYPED as local
function clearParty()
	if not plist then return end
	while #(plist.group) > 0 do
		local temp = tremove(plist.group)
		temp:Hide()
		plist:SetHeight(plist:GetHeight()-temp:GetHeight())
		recyclePlayer(temp)
	end
end

-- PROTOTYPED as local
function insertPlayer(player)
	-- Keep in mind, at this point "player" is an info table, not a frame
	if not plist then return end
	local frame = manufacturePlayerFrame(player)
	if #(plist.group) > 0 then
		-- Attach to the last player
		local last = plist.group[#(plist.group)]
		frame:SetPoint("TOPLEFT",last,"BOTTOMLEFT")
		frame:SetPoint("TOPRIGHT",last,"BOTTOMRIGHT")
	else
		-- Attach to the top of the player list
		frame:SetPoint("TOPLEFT",plist,"TOPLEFT")
		frame:SetPoint("TOPRIGHT",plist,"TOPRIGHT")
	end
	plist:SetHeight(plist:GetHeight()+frame:GetHeight())
	tinsert(plist.group,frame)
end

function DungeonMate:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("DungeonMateDB")
	self.db:RegisterDefaults({
		char = {
			["panel_position"] = {
				["point"] = "CENTER",
				["x"] = 450,
				["y"] = 0,
			}
		}
	})
	self:RegisterChatCommand("dm",showPanel)
	self:RegisterChatCommand("dungeonmate",showPanel)
end

function DungeonMate:OnEnable()
	initializeFrame()
	showPanel()
	DungeonMate:RegisterEvent("PARTY_MEMBERS_CHANGED",UpdatePartyDisplay)
	lgt:RegisterCallback("LibGroupTalents_Update",UpdatePartyDisplay) 
	DungeonMate:RegisterEvent("IGNORELIST_UPDATE",UpdatePartyDisplay)
end

function DungeonMate:OnDisable()
	hidePanel()
end



--[[ DATA BROKER ]]--

local ldb = LibStub("LibDataBroker-1.1")

ldb:NewDataObject("DungeonMate", {
	type = "launcher",
	icon = dmicon,
	OnClick = showPanel,
})
